@@ -17,7 +17,7 @@ import errno
 from http import client as httplib
 import multiprocessing
 import os
-import socket
+import queue
 import sys
 import threading
 import time
@@ -40,6 +40,7 @@ from kuryr_kubernetes import clients
 from kuryr_kubernetes.cni import handlers as h_cni
 from kuryr_kubernetes.cni import health
 from kuryr_kubernetes.cni.plugins import k8s_cni_registry
+from kuryr_kubernetes.cni import prometheus_exporter
 from kuryr_kubernetes.cni import utils as cni_utils
 from kuryr_kubernetes import config
 from kuryr_kubernetes import constants as k_const
@@ -57,10 +58,11 @@ ErrInternal = 999
 
 
 class DaemonServer(object):
-    def __init__(self, plugin, healthy):
+    def __init__(self, plugin, healthy, metrics):
         self.ctx = None
         self.plugin = plugin
         self.healthy = healthy
+        self.metrics = metrics
         self.failure_count = multiprocessing.Value('i', 0)
         self.application = flask.Flask('kuryr-daemon')
         self.application.add_url_rule(
@@ -86,6 +88,12 @@ class DaemonServer(object):
         data = jsonutils.dumps(template)
         return data
 
+    def _update_metrics(self, command, error, duration):
+        """Add a new metric value to the shared metrics dict"""
+        labels = {'command': command, 'error': error}
+        self.metrics.put({'labels': labels, 'duration': duration})
+
+    @cni_utils.measure_time('ADD')
     def add(self):
         try:
             params = self._prepare_request()
@@ -99,25 +107,20 @@ class DaemonServer(object):
         try:
             vif = self.plugin.add(params)
             data = jsonutils.dumps(vif.obj_to_primitive())
-        except exceptions.ResourceNotReady as e:
-            self._check_failure()
-            LOG.error('Error when processing addNetwork request')
-            error = self._error(ErrTryAgainLater,
-                                f"{e}. Try Again Later.")
+        except exceptions.CNITimeout as e:
+            LOG.exception('Timeout on ADD request')
+            error = self._error(ErrTryAgainLater, f"{e}. Try Again Later.")
             return error, httplib.GATEWAY_TIMEOUT, self.headers
         except pyroute2.NetlinkError as e:
             if e.code == errno.EEXIST:
                 self._check_failure()
-                args = {'kind': 'vlan', 'vlan_id': vif.vlan_id}
                 LOG.warning(
                     f'Creation of pod interface failed due to VLAN ID '
-                    f'(vlan_info={args}) conflict. Probably the CRI had not '
-                    f'cleaned up the network namespace of deleted pods. '
-                    f'Attempting to retry.')
+                    f'conflict. Probably the CRI had not cleaned up the '
+                    f'network namespace of deleted pods. Attempting to retry.')
                 error = self._error(ErrTryAgainLater,
-                                    "Creation of pod interface failed due to"
-                                    " vlan_id. Try Again Later",
-                                    f"vlan_id:{vif.vlan_id}")
+                                    "Creation of pod interface failed due to "
+                                    "VLAN ID conflict. Try Again Later")
                 return error, httplib.GATEWAY_TIMEOUT, self.headers
             raise
         except Exception:
@@ -138,6 +141,7 @@ class DaemonServer(object):
 
         return data, httplib.ACCEPTED, self.headers
 
+    @cni_utils.measure_time('DEL')
     def delete(self):
         try:
             params = self._prepare_request()
@@ -149,7 +153,7 @@ class DaemonServer(object):
 
         try:
             self.plugin.delete(params)
-        except exceptions.ResourceNotReady:
+        except exceptions.CNIKuryrPortTimeout:
             # NOTE(dulek): It's better to ignore this error - most of the time
             #              it will happen when pod is long gone and CRI
             #              overzealously tries to delete it from the network.
@@ -185,6 +189,12 @@ class DaemonServer(object):
             LOG.exception('Cannot start server on %s.', server_pair)
             raise
 
+        if CONF.cni_daemon.worker_num <= 1:
+            msg = ('[cni_daemon]worker_num needs to be set to a value higher '
+                   'than 1')
+            LOG.critical(msg)
+            raise exceptions.InvalidKuryrConfiguration(msg)
+
         try:
             self._server = serving.make_server(
                 address, port, self.application, threaded=False,
@@ -215,13 +225,14 @@ class DaemonServer(object):
 class CNIDaemonServerService(cotyledon.Service):
     name = "server"
 
-    def __init__(self, worker_id, registry, healthy):
+    def __init__(self, worker_id, registry, healthy, metrics):
         super(CNIDaemonServerService, self).__init__(worker_id)
         self.registry = registry
         self.healthy = healthy
         self.plugin = k8s_cni_registry.K8sCNIRegistryPlugin(registry,
                                                             self.healthy)
-        self.server = DaemonServer(self.plugin, self.healthy)
+        self.metrics = metrics
+        self.server = DaemonServer(self.plugin, self.healthy, self.metrics)
 
     def run(self):
         # NOTE(dulek): We might do a *lot* of pyroute2 operations, let's
@@ -247,23 +258,13 @@ class CNIDaemonWatcherService(cotyledon.Service):
         self.registry = registry
         self.healthy = healthy
 
-    def _get_nodename(self):
-        # NOTE(dulek): At first try to get it using environment variable,
-        #              otherwise assume hostname is the nodename.
-        try:
-            nodename = os.environ['KUBERNETES_NODE_NAME']
-        except KeyError:
-            # NOTE(dulek): By default K8s nodeName is lowercased hostname.
-            nodename = socket.gethostname().lower()
-        return nodename
-
     def run(self):
         self.pipeline = h_cni.CNIPipeline()
         self.pipeline.register(h_cni.CallbackHandler(self.on_done,
                                                      self.on_deleted))
         self.watcher = k_watcher.Watcher(self.pipeline)
         query_label = urllib.parse.quote_plus(f'{k_const.KURYRPORT_LABEL}='
-                                              f'{self._get_nodename()}')
+                                              f'{utils.get_nodename()}')
 
         self.watcher.add(f'{k_const.K8S_API_CRD_KURYRPORTS}'
                          f'?labelSelector={query_label}')
@@ -340,11 +341,43 @@ class CNIDaemonHealthServerService(cotyledon.Service):
         self.health_server.run()
 
 
+class CNIDaemonExporterService(cotyledon.Service):
+    name = "Prometheus Exporter"
+
+    def __init__(self, worker_id, metrics):
+        super(CNIDaemonExporterService, self).__init__(worker_id)
+        self.prometheus_exporter = prometheus_exporter.CNIPrometheusExporter()
+        self.is_running = True
+        self.metrics = metrics
+        self.exporter_thread = threading.Thread(
+            target=self._start_metric_updater)
+        self.exporter_thread.start()
+
+    def _start_metric_updater(self):
+        while self.is_running:
+            try:
+                metric = self.metrics.get(timeout=1)
+            except queue.Empty:
+                continue
+            labels = metric['labels']
+            duration = metric['duration']
+            self.prometheus_exporter.update_metric(labels, duration)
+
+    def terminate(self):
+        self.is_running = False
+        if self.exporter_thread:
+            self.exporter_thread.join()
+
+    def run(self):
+        self.prometheus_exporter.run()
+
+
 class CNIDaemonServiceManager(cotyledon.ServiceManager):
     def __init__(self):
         # NOTE(mdulko): Default shutdown timeout is 60 seconds and K8s won't
         #               wait more by default anyway.
         super(CNIDaemonServiceManager, self).__init__()
+        self._server_service = None
         # TODO(dulek): Use cotyledon.oslo_config_glue to support conf reload.
 
         # TODO(vikasc): Should be done using dynamically loadable OVO types
@@ -359,11 +392,20 @@ class CNIDaemonServiceManager(cotyledon.ServiceManager):
         self.manager = multiprocessing.Manager()
         registry = self.manager.dict()  # For Watcher->Server communication.
         healthy = multiprocessing.Value(c_bool, True)
+        metrics = self.manager.Queue()
         self.add(CNIDaemonWatcherService, workers=1, args=(registry, healthy,))
-        self._server_service = self.add(
-            CNIDaemonServerService, workers=1, args=(registry, healthy))
+        self._server_service = self.add(CNIDaemonServerService, workers=1,
+                                        args=(registry, healthy, metrics,))
         self.add(CNIDaemonHealthServerService, workers=1, args=(healthy,))
-        self.register_hooks(on_terminate=self.terminate)
+        self.add(CNIDaemonExporterService, workers=1, args=(metrics,))
+
+        def shutdown_hook(service_id, worker_id, exit_code):
+            LOG.critical(f'Child Service {service_id} had exited with code '
+                         f'{exit_code}, stopping kuryr-daemon')
+            self.shutdown()
+
+        self.register_hooks(on_terminate=self.terminate,
+                            on_dead_worker=shutdown_hook)
 
     def run(self):
         # FIXME(darshna): Remove pyroute2 IPDB deprecation warning, remove
@@ -396,12 +438,13 @@ class CNIDaemonServiceManager(cotyledon.ServiceManager):
 
     def terminate(self):
         self._terminate_called.set()
-        LOG.info("Gracefully stopping DaemonServer service..")
-        self.reconfigure(self._server_service, 0)
-        for worker in self._running_services[self._server_service]:
-            worker.terminate()
-        for worker in self._running_services[self._server_service]:
-            worker.join()
+        if self._server_service:
+            LOG.info("Gracefully stopping DaemonServer service..")
+            self.reconfigure(self._server_service, 0)
+            for worker in self._running_services[self._server_service]:
+                worker.terminate()
+            for worker in self._running_services[self._server_service]:
+                worker.join()
         LOG.info("Stopping registry manager...")
         self.manager.shutdown()
         LOG.info("Continuing with shutdown")
