@@ -11,6 +11,7 @@
 # under the License.
 
 import ipaddress
+import os
 import random
 import re
 import socket
@@ -91,6 +92,9 @@ RESOURCE_MAP = {'Endpoints': 'endpoints',
                 'Pod': 'pods',
                 'Service': 'services',
                 'Machine': 'machines'}
+API_VER_MAP = {'NetworkPolicy': 'networking.k8s.io/v1',
+               'Pod': 'v1',
+               'Service': 'v1'}
 API_RE = re.compile(r'v\d+')
 
 
@@ -636,3 +640,80 @@ def get_subnet_by_ip(nodes_subnets, target_ip):
             return nodes_subnet
 
     return None
+
+
+def get_kuryrloadbalancer(name, namespace):
+    k8s = clients.get_kubernetes_client()
+    try:
+        return k8s.get(f'{constants.K8S_API_CRD_NAMESPACES}/'
+                       f'{namespace}/kuryrloadbalancers/'
+                       f'{name}')
+    except exceptions.K8sResourceNotFound:
+        return {}
+
+
+def is_pod_completed(pod):
+    try:
+        return (pod['status']['phase'] in
+                (constants.K8S_POD_STATUS_SUCCEEDED,
+                 constants.K8S_POD_STATUS_FAILED))
+    except KeyError:
+        return False
+
+
+def is_host_network(pod):
+    return pod['spec'].get('hostNetwork', False)
+
+
+def get_nodename():
+    # NOTE(dulek): At first try to get it using environment variable,
+    #              otherwise assume hostname is the nodename.
+    try:
+        nodename = os.environ['KUBERNETES_NODE_NAME']
+    except KeyError:
+        # NOTE(dulek): By default K8s nodeName is lowercased hostname.
+        nodename = socket.gethostname().lower()
+    return nodename
+
+
+def get_referenced_object(obj, kind):
+    """Get referenced object.
+
+    Helper function for getting objects out of the CRDs like
+    KuryrLoadBalancer, KuryrNetworkPolicy or KuryrPort needed solely for
+    creating Event object, so there will be no exceptions raises from this
+    function.
+    """
+    for ref in obj['metadata'].get('ownerReferences', []):
+        if ref['kind'] != kind:
+            continue
+
+        try:
+            return {'kind': kind,
+                    'apiVersion': ref['apiVersion'],
+                    'metadata': {'namespace': obj['metadata']['namespace'],
+                                 'name': ref['name'],
+                                 'uid': ref['uid']}}
+        except KeyError:
+            LOG.debug("Not all needed keys was found in ownerReferences "
+                      "list: %s", ref)
+
+    # There was no ownerReferences field, let's query API
+    k8s = clients.get_kubernetes_client()
+    data = {'metadata': {'name': obj['metadata']['name']},
+            'kind': kind,
+            'apiVersion': API_VER_MAP[kind]}
+    if obj['metadata'].get('namespace'):
+        data['metadata']['namespace'] = obj['metadata']['namespace']
+    try:
+        url = get_res_link(data)
+    except KeyError:
+        LOG.debug("Not all needed data was found in provided object: %s",
+                  data)
+        return
+
+    try:
+        return k8s.get(url)
+    except exceptions.K8sClientException:
+        LOG.debug('Error when fetching %s to add an event %s, ignoring',
+                  kind, get_res_unique_name(obj))
